@@ -1,17 +1,51 @@
-import asyncio
-import contextlib
 import enum
-import inspect
 import io
+import logging
 import re
 from datetime import datetime
-from typing import Any, Optional, Union
+from functools import partial
+from typing import Any, AsyncIterable, Callable, Coroutine, Dict, Optional, Union
 
+import anyio
 from starlette.background import BackgroundTask
-from starlette.concurrency import iterate_in_threadpool, run_until_first_complete
+from starlette.concurrency import iterate_in_threadpool
 from starlette.responses import Response
 from starlette.types import Receive, Scope, Send
-from uvicorn.config import logger as _log
+
+_log = logging.getLogger(__name__)
+
+
+# https://stackoverflow.com/questions/58133694/graceful-shutdown-of-uvicorn-starlette-app-with-websockets
+class AppStatus:
+    """helper for monkey-patching the signal-handler of uvicorn"""
+
+    should_exit = False
+    should_exit_event: Union[anyio.Event, None] = None
+
+    @staticmethod
+    def handle_exit(*args, **kwargs):
+        # set bool flag before checking the event to avoid race condition
+        AppStatus.should_exit = True
+        # Check if event has been initialized, if so notify listeners
+        if AppStatus.should_exit_event is not None:
+            AppStatus.should_exit_event.set()
+        original_handler(*args, **kwargs)
+
+
+try:
+    from uvicorn.main import Server
+
+    original_handler = Server.handle_exit
+    Server.handle_exit = AppStatus.handle_exit  # type: ignore
+
+    def unpatch_uvicorn_signal_handler():
+        """restores original signal-handler and rolls back monkey-patching.
+        Normally this should not be necessary.
+        """
+        Server.handle_exit = original_handler
+
+except ModuleNotFoundError:
+    _log.debug("Uvicorn not used.")
 
 
 class SseState(enum.Enum):
@@ -23,23 +57,14 @@ class SseState(enum.Enum):
 class ServerSentEvent:
     def __init__(
         self,
-        data: Any,
+        data: Optional[Any] = None,
         *,
         event: Optional[str] = None,
         id: Optional[int] = None,
         retry: Optional[int] = None,
-        sep: str = None,
+        comment: Optional[str] = None,
+        sep: Optional[str] = None,
     ) -> None:
-        self.data = data
-        self.event = event
-        self.id = id
-        self.retry = retry
-
-        self.DEFAULT_SEPARATOR = "\r\n"
-        self.LINE_SEP_EXPR = re.compile(r"\r\n|\r|\n")
-        self._sep = sep if sep is not None else self.DEFAULT_SEPARATOR
-
-    def encode(self) -> bytes:
         """Send data using EventSource protocol
 
         :param str data: The data field for the message.
@@ -53,8 +78,27 @@ class ServerSentEvent:
             the event. [What code handles this?] This must be an integer,
             specifying the reconnection time in milliseconds. If a non-integer
             value is specified, the field is ignored.
+        :param str comment: A colon as the first character of a line is essence
+            a comment, and is ignored. Usually used as a ping message to keep connecting.
+            If set, this will be a comment message.
         """
+        self.data = data
+        self.event = event
+        self.id = id
+        self.retry = retry
+        self.comment = comment
+        self.DEFAULT_SEPARATOR = "\r\n"
+        self.LINE_SEP_EXPR = re.compile(r"\r\n|\r|\n")
+        self._sep = sep if sep is not None else self.DEFAULT_SEPARATOR
+
+    def encode(self) -> bytes:
         buffer = io.StringIO()
+        if self.comment is not None:
+            for chunk in self.LINE_SEP_EXPR.split(str(self.comment)):
+                buffer.write(f": {chunk}")
+                buffer.write(self._sep)
+            return buffer.getvalue().encode("utf-8")
+
         if self.id is not None:
             buffer.write(self.LINE_SEP_EXPR.sub("", f"id: {self.id}"))
             buffer.write(self._sep)
@@ -77,57 +121,67 @@ class ServerSentEvent:
         return buffer.getvalue().encode("utf-8")
 
 
-class EventSourceResponse(Response):
-    """Implements the ServerSentEvent Protocol: https://www.w3.org/TR/2009/WD-eventsource-20090421/
+def ensure_bytes(data: Union[bytes, dict, ServerSentEvent, Any]) -> bytes:
+    if isinstance(data, bytes):
+        return data
+    elif isinstance(data, ServerSentEvent):
+        return data.encode()
+    elif isinstance(data, dict):
+        return ServerSentEvent(**data).encode()
+    else:
+        return ServerSentEvent(str(data)).encode()
 
-    Responses must not be compressed by middleware in order to work properly.
+
+class EventSourceResponse(Response):
+    """Implements the ServerSentEvent Protocol:
+    https://www.w3.org/TR/2009/WD-eventsource-20090421/
+
+    Responses must not be compressed by middleware in order to work.
+    implementation based on Starlette StreamingResponse
     """
 
     DEFAULT_PING_INTERVAL = 15
 
+    # noinspection PyMissingConstructor
     def __init__(
         self,
         content: Any,
-        # content: Iterator[Any],
-        # content: Union[
-        #     Generator[Union[str, Dict], None, None],
-        #     AsyncGenerator[Union[str, Dict], None],
-        # ],
         status_code: int = 200,
-        headers: dict = None,
-        media_type: str = "text/html",
-        background: BackgroundTask = None,
-        ping: int = None,
-        sep: str = None,
+        headers: Optional[Dict] = None,
+        media_type: str = "text/event-stream",
+        background: Optional[BackgroundTask] = None,
+        ping: Optional[int] = None,
+        sep: Optional[str] = None,
+        ping_message_factory: Optional[Callable[[], ServerSentEvent]] = None,
     ) -> None:
         self.sep = sep
-        if inspect.isasyncgen(content):
-            self.body_iterator = content
+        self.ping_message_factory = ping_message_factory
+        if isinstance(content, AsyncIterable):
+            self.body_iterator = (
+                content
+            )  # type: AsyncIterable[Union[Any,dict,ServerSentEvent]]
         else:
-            self.body_iterator = iterate_in_threadpool(content)
+            self.body_iterator = iterate_in_threadpool(content)  # type: ignore
         self.status_code = status_code
         self.media_type = self.media_type if media_type is None else media_type
-        self.background = background
+        self.background = background  # type: ignore  # follows https://github.com/encode/starlette/blob/master/starlette/responses.py
 
-        _headers = dict()
+        _headers = {}
         if headers is not None:  # pragma: no cover
             _headers.update(headers)
 
         # mandatory for servers-sent events headers
-        _headers["Content-Type"] = "text/event-stream"
-        _headers["Cache-Control"] = "no-cache"
+        # allow cache control header to be set by user to support fan out proxies
+        # https://www.fastly.com/blog/server-sent-events-fastly
+        _headers.setdefault("Cache-Control", "no-cache")
         _headers["Connection"] = "keep-alive"
         _headers["X-Accel-Buffering"] = "no"
-        # _headers['Transfer-Encoding'] = 'chunked'
 
         self.init_headers(_headers)
 
-        self._loop = None
         self.ping_interval = self.DEFAULT_PING_INTERVAL if ping is None else ping
-        self._ping_task = None
         self.active = True
 
-        self._loop = asyncio.get_event_loop()
         self._ping_task = None
 
     @staticmethod
@@ -135,21 +189,24 @@ class EventSourceResponse(Response):
         while True:
             message = await receive()
             if message["type"] == "http.disconnect":
-                _log.debug(f"Got event: http.disconnect. Stop streaming.")
+                _log.debug("Got event: http.disconnect. Stop streaming.")
                 break
 
-    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
-        await run_until_first_complete(
-            (self.stream_response, {"send": send}),
-            (self.listen_for_disconnect, {"receive": receive}),
-        )
+    @staticmethod
+    async def listen_for_exit_signal() -> None:
+        # Check if should_exit was set before anybody started waiting
+        if AppStatus.should_exit:
+            return
 
-        self.stop_streaming()
-        await self.wait()
-        _log.debug(f"streaming stopped.")
+        # Setup an Event
+        AppStatus.should_exit_event = anyio.Event()
 
-        if self.background is not None:  # pragma: no cover, tested in StreamResponse
-            await self.background()
+        # Check if should_exit got set while we set up the event
+        if AppStatus.should_exit:
+            return
+
+        # Await the event
+        await AppStatus.should_exit_event.wait()
 
     async def stream_response(self, send) -> None:
         await send(
@@ -159,34 +216,28 @@ class EventSourceResponse(Response):
                 "headers": self.raw_headers,
             }
         )
-        self._ping_task = self._loop.create_task(self._ping(send))  # type: ignore
         async for data in self.body_iterator:
-            if isinstance(data, dict):
-                chunk = ServerSentEvent(**data).encode()
-            else:
-                chunk = ServerSentEvent(str(data), sep=self.sep).encode()
+            chunk = ensure_bytes(data)
             _log.debug(f"chunk: {chunk.decode()}")
             await send({"type": "http.response.body", "body": chunk, "more_body": True})
+
         await send({"type": "http.response.body", "body": b"", "more_body": False})
 
-    async def wait(self) -> None:
-        """EventSourceResponse object is used for streaming data to the client,
-        this method returns future, so we can wait until connection will
-        be closed or other task explicitly call ``stop_streaming`` method.
-        """
-        if self._ping_task is None:
-            raise RuntimeError("Response is not started")
-        with contextlib.suppress(asyncio.CancelledError):
-            await self._ping_task
-            _log.debug(f"SSE ping stopped.")  # pragma: no cover
+    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
+        async with anyio.create_task_group() as task_group:
+            # https://trio.readthedocs.io/en/latest/reference-core.html#custom-supervisors
+            async def wrap(func: Callable[[], Coroutine[None, None, None]]) -> None:
+                await func()
+                # noinspection PyAsyncCall
+                task_group.cancel_scope.cancel()
 
-    def stop_streaming(self) -> None:
-        """Used in conjunction with ``wait`` could be called from other task
-        to notify client that server no longer wants to stream anything.
-        """
-        if self._ping_task is None:
-            raise RuntimeError("Response is not started")
-        self._ping_task.cancel()
+            task_group.start_soon(wrap, partial(self.stream_response, send))
+            task_group.start_soon(wrap, partial(self._ping, send))
+            task_group.start_soon(wrap, self.listen_for_exit_signal)
+            await wrap(partial(self.listen_for_disconnect, receive))
+
+        if self.background is not None:  # pragma: no cover, tested in StreamResponse
+            await self.background()
 
     def enable_compression(self, force: bool = False) -> None:
         raise NotImplementedError
@@ -217,7 +268,13 @@ class EventSourceResponse(Response):
         # Alternatively one can send periodically a comment line
         # (one starting with a ':' character)
         while self.active:
-            await asyncio.sleep(self._ping_interval)
-            ping = ServerSentEvent(datetime.utcnow(), event="ping").encode()
+            await anyio.sleep(self._ping_interval)
+            if self.ping_message_factory:
+                assert isinstance(self.ping_message_factory, Callable)  # type: ignore  # https://github.com/python/mypy/issues/6864
+            ping = (
+                ServerSentEvent(datetime.utcnow(), event="ping").encode()
+                if self.ping_message_factory is None
+                else ensure_bytes(self.ping_message_factory())
+            )
             _log.debug(f"ping: {ping.decode()}")
             await send({"type": "http.response.body", "body": ping, "more_body": True})
